
/*-------------------------------------------------------------------------------*\
  statlink: Monitoring the status of PCIe40 board and belle2links
 
  Qi-Dong Zhou, KEK 

  2020.04.28  0.01 first version based on staths.c
  2020.12.19  0.01 rename statlink to pcie40_statlink
  2020.12.23  0.02 add protection for pll status access
  2021.01.07  0.03 add more monitoring info
\*--------------------------------------------------------------------------------*/
#define VERSION 03
#define MOD_DATE 20210107

#include "pcie40_b2slc.h"
#include "pcie40_reg.h"
#include "pcie40_b2config.h"
#include "pcie40_ecs.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <iostream>
#include <vector>
#include <bitset>
#include <sys/file.h>

#define MAX_NUM_CH   48
#define MAX_SLOT     3
#define SLC_BAR      2
#define SIZE_MEMO_FPGA 1024 // maximum size of memory in FPGA

bool VARBOSE = false;
bool FEE     = false;

#ifndef D
#define D(a,b,c) (((a)>>(c))&((1<<((b)+1-(c)))-1))
#define B(a,b)   D(a,b,b)
#define Bs(a,b,s)   (B(a,b)?(s):"")
#define Ds(a,b,c,s)   (D(a,b,c)?(s):"")
#endif

// parse 
int dev_slot         = 0; // default device slot number, for use multiple PCIe40 on a server,--dev xx option
int ch               = -1;
std::bitset<48> link_mask; 
//Flag for switch slc access method

void argument(int argc, char **argv){
  char *ARGV0 = argv[0];
  std::bitset<48> en_link;
  if(argc == 1){
    link_mask.set();
    //std::cout<<std::hex<<link_mask<<std::endl;
  }

  for(int i=1;i<argc;i++){
    std::string ss = argv[i];
    if(ss=="--dev"){
      dev_slot = atoi(argv[++i]);
      if( dev_slot<0 || dev_slot >= MAX_SLOT){
	fprintf(stderr, "Invalid device slot %d, Valid slot [0, %d]\n", dev_slot, MAX_SLOT);
	exit(1);
      }
    }
    if(ss=="--varbose"){
      VARBOSE = true;
    }
    if(ss=="--fee"){
      FEE = true;
    }
    if(ss=="--ch"){
      for(int j = i+1; j<argc; j++){
	std::string sss = argv[j];
	if(sss == "--varbose")break;
	if(sss == "--fee")break;
	if(sss == "all"){
	  link_mask.set();	
	  break;	
	}
	ch = atoi(argv[j]);
	if( ch<0 || ch >= MAX_NUM_CH){
	  fprintf(stderr, "Invalid channel ID %d, Valid ID [0, %d]\n", ch, MAX_NUM_CH);
	  exit(1);
	}
	en_link.reset();
	en_link.set(ch, true);
	link_mask = link_mask | en_link;
	//std::cout <<i << " " << j <<" "<< ch <<" "<<std::hex<< en_link <<" "<<std::hex<<link_mask<< std::endl;
	
      }
    }
    if(ss=="-h" || ss=="-help"){
      fprintf(stderr, "%s version %d.%02d date %d\n"
            , ARGV0, VERSION/100, VERSION%100, MOD_DATE);
      fprintf(stderr,  "usage: %s --ch 0 \n"
                       "       --dev   xx              #device slot number which installed PCIe40\n" 
	               "       --ch    xx  xx ...      #link channel number\n" 
                       , ARGV0); 
      exit(1);
    }
  }
}

/* ---------------------------------------------------------------------- *\
 *    getfee
\* ---------------------------------------------------------------------- */
int
getfee( pcie40reg_t *pcie40p)
{
  int hwtype;
  int serial;
  int fwtype;
  int fwver;

  hwtype = pcie40_readfee8(dev_slot, ch, PCIE40REG_FEEHWTYPE);
  serial = pcie40_readfee8(dev_slot, ch, PCIE40REG_FEESERIAL);
  fwtype = pcie40_readfee8(dev_slot, ch, PCIE40REG_FEEFWTYPE);
  fwver  = pcie40_readfee8(dev_slot, ch, PCIE40REG_FEEFWVER);

  pcie40p->feeser = (serial | (hwtype << 8)) & 0xfff;
  pcie40p->feever = (fwver  | (fwtype << 8)) & 0xfff;
  pcie40p->feehw  = (hwtype >> 4) & 0x0f;
  pcie40p->feefw  = (fwtype >> 4) & 0x0f;
  
  pcie40p->feecrce = pcie40_readfee8(dev_slot, ch, PCIE40REG_CRCERR);

  return 0;
}

/* ---------------------------------------------------------------------- *\
 *    readregs
\* ---------------------------------------------------------------------- */
int
readregs(std::bitset<48> link_mask, pcie40reg_t pcie40reg[])
{
  int i;
  int notfound = 1;

  for (i=0; i<48; i++) {
    pcie40reg_t *pcie = &pcie40reg[i];
    memset(pcie, 0, sizeof(*pcie));
    if (!link_mask[i]) continue;
    
    pcie->eventsz    = pcie40_b2l_evtsz(dev_slot, i);  
    pcie->nevent     = pcie40_b2l_nevent(dev_slot, i);
    pcie->nkbyte     = pcie40_b2l_nkyte(dev_slot, i);  
    pcie->maxevtsz   = pcie40_b2l_maxevtsz(dev_slot, i);
    pcie->ffusage    = pcie40_b2l_ffusage(dev_slot, i);
    pcie->lffusage   = pcie40_b2l_lffusage(dev_slot, i);
    pcie->rstgbt     = pcie40_b2l_rstgbt(dev_slot, i);
    pcie->cntcrc32   = pcie40_b2l_cntcrc32(dev_slot, i);
    pcie->cflagOK    = pcie40_b2l_checkFlagOK(dev_slot, i);
    pcie->cflagNG    = pcie40_b2l_checkFlagNG(dev_slot, i);
    pcie->monstatus  = pcie40_b2l_monstatus(dev_slot, i);
    pcie->vetoidle   = pcie40_b2l_cnt_vetoidle(dev_slot, i);
    pcie->vetodata   = pcie40_b2l_cnt_vetodata(dev_slot, i);
    pcie->vetobuf[0] = pcie40_b2l_bufdata0(dev_slot, i);
    pcie->vetobuf[1] = pcie40_b2l_bufdata1(dev_slot, i);
    if(i>=0 && i<24){
      pcie->b2lsta[0]   = pcie40_b2l_status(dev_slot, i);
      pcie->b2lgbt[0]   = pcie40_b2l_gbt(dev_slot, i);
      pcie->b2ltxsta[0] = pcie40_b2l_txstatus(dev_slot, i);
      pcie->b2ltxrdy[0] = pcie40_b2l_txready(dev_slot, i);
      pcie->b2lrxsta[0] = pcie40_b2l_rxstatus(dev_slot, i);
      pcie->b2lrxrdy[0] = pcie40_b2l_rxready(dev_slot, i);
      pcie->b2lmask[0]  = pcie40_b2l_mask(dev_slot, i);
      pcie->b2lfful[0]  = pcie40_b2l_fffull(dev_slot, i);
    }else if(i>=24 && i<48){
      pcie->b2lsta[1]   = pcie40_b2l_status(dev_slot, i);
      pcie->b2lgbt[1]   = pcie40_b2l_gbt(dev_slot, i);
      pcie->b2ltxsta[1] = pcie40_b2l_txstatus(dev_slot, i);
      pcie->b2ltxrdy[1] = pcie40_b2l_txready(dev_slot, i);
      pcie->b2lrxsta[1] = pcie40_b2l_rxstatus(dev_slot, i);
      pcie->b2lrxrdy[1] = pcie40_b2l_rxready(dev_slot, i);
      pcie->b2lmask[1]  = pcie40_b2l_mask(dev_slot, i);
      pcie->b2lfful[1]  = pcie40_b2l_fffull(dev_slot, i);
    }
    
    pcie->feeser = 0xffff;
    pcie->feever = 0xffff;
    if ( pcie40_b2l_status(dev_slot, i) && FEE) {
      getfee(pcie);
    }
  }
  return notfound;
}


/* ---------------------------------------------------------------------- *\
 *    statpcie40
\* ---------------------------------------------------------------------- */
int
statpice40(pcie40reg_t pcie40reg)
{
  FILE *fp = fopen("/tmp/pcie40_pll_lockfile.lck","w");
  pcie40reg_t *pcie = &pcie40reg;
  memset(pcie, 0, sizeof(*pcie));
  
  pcie->status     = pcie40_status(dev_slot);
  pcie->nkbyte2dma = pcie40_nkbyte_dma(dev_slot);
  pcie->dmastatus  = pcie40_dma_status(dev_slot);
  std::cout << "PCIE40 firmware version " << pcie40_fpgaVersion(dev_slot) << std::endl;
  printf("memory: %s | ttd: %s |  ttd clk: %s | run=: %d | trg: %d | trg type: %d\n", 
	 B(pcie->status, 8)?"FULL":"OK",
	 B(pcie->status, 12)?"UP":"DOWN",
	 B(pcie->status, 11)?"UP":"DOWN",
	 pcie40_run_number(dev_slot),
	 D(pcie->status, 28, 17),
	 pcie40_trg_type(dev_slot));	 
  flock( fileno(fp), LOCK_EX );
  printf("PLLs:%s | B2L:%s (rx:%d%d%d tx:%d%d) | DMA:%s (%9.1fkB)\n", 	  
	 (pcie40_pllLockStatus(dev_slot, 1)&&pcie40_pllLockStatus(dev_slot, 2)&&pcie40_pllLockStatus(dev_slot, 3))?"LOCKED":"NOT LOCKED",
	 (B(pcie->status, 0) && B(pcie->status, 1) && B(pcie->status, 2) && !B(pcie->status, 3) && !B(pcie->status, 5))?"READY":"NOT READY",
	 B(pcie->status, 0),
	 !B(pcie->status, 3),
	 !B(pcie->status, 4),
	 B(pcie->status, 1),
	 B(pcie->status, 2),
	 B(pcie->dmastatus,0)?"FREE":"FULL",
         pcie->nkbyte2dma);

  if(VARBOSE){
    printf("--------------------------------\n");
    printf("Input frequency = %f [MHz]\n",(double)pcie40_inputPLLFrequency(dev_slot)/1000000);
    printf("--------\n");
    printf("PLL No.   :   Lock sta | Input sta | Fre. sta | Lock cnt | Input. cnt | Fre. cnt\n");
    printf("PLL[0-23] : %*s | %s | %s | %s | %s | %s\n", 10,
	   pcie40_pllLockStatus(dev_slot, 1)?"LOCKED":"NOT LOCKED",
	   pcie40_pllInputStatus(dev_slot, 1)?"OK ":"BAD",
	   pcie40_pllFrequencyStatus(dev_slot, 1)?"OK ":"BAD",
	   pcie40_pllLockCounter(dev_slot, 1)?"OK ":"LOL",
	   pcie40_pllInputCounter(dev_slot, 1)?"OK ":"LOS",
	   pcie40_pllFrequencyCounter(dev_slot, 1)?"OK ":"OOF");
    printf("PLL[24-47]: %*s | %s | %s | %s | %s | %s\n", 10,
	   pcie40_pllLockStatus(dev_slot, 2)?"LOCKED":"NOT LOCKED",
	   pcie40_pllInputStatus(dev_slot, 2)?"OK ":"BAD",
	   pcie40_pllFrequencyStatus(dev_slot, 2)?"OK ":"BAD",
	   pcie40_pllLockCounter(dev_slot, 2)?"OK ":"LOL",
	   pcie40_pllInputCounter(dev_slot, 2)?"OK ":"LOS",
	   pcie40_pllFrequencyCounter(dev_slot, 2)?"OK ":"OOF");
    printf("PLL[FTSW] : %*s | %s | %s | %s | %s | %s\n", 10,
	   pcie40_pllLockStatus(dev_slot, 3)?"LOCKED":"NOT LOCKED",
	   pcie40_pllInputStatus(dev_slot, 3)?"OK ":"BAD",
	   pcie40_pllFrequencyStatus(dev_slot, 3)?"OK ":"BAD",
	   pcie40_pllLockCounter(dev_slot, 3)?"OK ":"LOL",
	   pcie40_pllInputCounter(dev_slot, 3)?"OK ":"LOS",
	   pcie40_pllFrequencyCounter(dev_slot, 3)?"OK ":"OOF");
    printf("--------------------------------\n");
  }
  flock( fileno(fp), LOCK_UN );
  fclose(fp);
  return 0;
}


/* ---------------------------------------------------------------------- *\
 *    statlink
\* ---------------------------------------------------------------------- */
int
statlink(std::bitset<48> link_mask, pcie40reg_t pcie40reg[])
{
  int i;
  int first = 1;
  int quiet = 1;

  for (i=0; i<48; i++) {
    pcie40reg_t *pcie = &pcie40reg[i];
    double total;
    char prompt[128];
    
    if (!link_mask[i]) continue;
    
    quiet = 0;
    if (! first) printf("\n");
    first =0;
    
    sprintf(prompt, "(%02d)    ", i);
    
    if ( ! pcie40_b2l_status(dev_slot, i) ) {
      printf("b2link-%02d is down\n", i);
    }else if(!FEE){
      printf("No fee flag is specified\n");
    }else if (pcie->feeser & 0x8000) {
      printf("fee info is not available\n");
    } else {
      printf("%s serial %d version %d\n",
	     feename(pcie->feehw, pcie->feefw), pcie->feeser, pcie->feever);
    }
    
    
    printf("%s", prompt);
    if(i>=0 && i<24){
    printf("b2l=%s (gbt=%s rx=%s tx=%s rxsta=%s txsta=%s mask=%s)\n",
	   pcie->b2lsta[0] ?"UP":"DOWN",
	   pcie->b2lgbt[0] ?"UP":"DOWN",
	   pcie->b2lrxsta[0] ?"UP":"DOWN",
	   pcie->b2ltxsta[0] ?"UP":"DOWN",
	   pcie->b2lrxrdy[0] ?"READY":"NOT READY",
	   pcie->b2ltxrdy[0] ?"READY":"NOT READY",
	   pcie->b2lmask[0] ?"UNMASK":"MASK");
    }else if (i>=24 && i<48){
          printf("b2l=%s (gbt=%s rx=%s tx=%s rxsta=%s txsta=%s mask=%s)\n",
	   pcie->b2lsta[1] ?"UP":"DOWN",
	   pcie->b2lgbt[1] ?"UP":"DOWN",
	   pcie->b2lrxsta[1]?"UP":"DOWN",
	   pcie->b2ltxsta[1]?"UP":"DOWN",
	   pcie->b2lrxrdy[1]?"READY":"NOT READY",
	   pcie->b2ltxrdy[1]?"READY":"NOT READY",
	   pcie->b2lmask[1]?"UNMASK":"MASK");

    }
    printf("%s", prompt);

    printf("event=%-9d total=%8.1fkB", pcie->nevent, pcie->nkbyte); 
    if (pcie->nevent) {
      double avg = pcie->nkbyte / pcie->nevent;
      printf(" (avg=%4.3fkB", avg);
      if (pcie->eventsz == 0xffff) {
        printf(" last=oflow");
      }else{
        printf(" last=%4.3fkB", pcie->eventsz);
      }
      if (pcie->maxevtsz == 0xffff) {
        printf(" max=oflow)");
      } else {
        printf(" max=%4.3fkB)", pcie->maxevtsz);
      }
    }
    printf("\n");
    

    printf("%s", prompt);
    if(i>=0 && i<24){
      printf("full=%d ",pcie->b2lfful[0]);
      if(pcie->b2lfful[0]){
	printf("ff=%d lenff=%d ", pcie->ffusage, pcie->lffusage);
      }
    }else if(i>=24 && i<48){
      printf("full=%d ",pcie->b2lfful[1]);
      if(pcie->b2lfful[1]){
	printf("ff=%d lenff=%d ", pcie->ffusage, pcie->lffusage);
      }
    }
    if(FEE){
      printf("feecrcerr=%d ", pcie->feecrce);
    }
    printf("check=%s ",
	   (pcie->cntcrc32==0 && pcie->cflagOK==0x3E && pcie->cflagNG==0)?"OK":"NG");
    if(!(pcie->cntcrc32==0 && pcie->cflagOK==0x3E && pcie->cflagNG==0)){
      printf("rxcrcerr=%d, check=%x %s %s %s %s %s", pcie->cntcrc32, pcie->cflagNG,
	     Bs(pcie->cflagNG,1,"crc16"),
	     Bs(pcie->cflagNG,2,"tt-tag"),
	     Bs(pcie->cflagNG,3,"tt-tagX"),
	     Bs(pcie->cflagNG,4,"tt-ctime"),
	     Bs(pcie->cflagNG,5,"exprun"));
    }      
    printf("\n");
    
    printf("%s", prompt);
    
    if ( pcie->monstatus & 6 == 0) {
      printf("no b2link error correction");
    } else if (pcie->vetoidle == 0 && pcie->vetodata == 0) {
      printf("no b2link error");
    } else {
      printf("b2link error %d(%c) %d(%c) %c%02x %c%02x %c%02x %c%02x %c%02x %c%02x",
	     pcie->vetoidle,
	     B(pcie->monstatus, 3) ? 'i' : '-',
	     pcie->vetodata,
	     B(pcie->monstatus, 2) ? 'd' : '-',
	     B(pcie->vetobuf[1], 21) ? 'K':'D', D(pcie->vetobuf[1],15,8),
	     B(pcie->vetobuf[1], 20) ? 'K':'D', D(pcie->vetobuf[1],7,0),
	     B(pcie->vetobuf[1], 19) ? 'K':'D', D(pcie->vetobuf[0],31,24),
	     B(pcie->vetobuf[1], 18) ? 'K':'D', D(pcie->vetobuf[0],23,16),
	     B(pcie->vetobuf[1], 17) ? 'K':'D', D(pcie->vetobuf[0],15,8),
	     B(pcie->vetobuf[1], 16) ? 'K':'D', D(pcie->vetobuf[0],7,0));
    }
    printf("\n");
  }
    

  if (quiet) {
    printf("no B2LINK No. was found\n");
  }
  return 0 ;
}


int main(int argc, char** argv){
  
  // parse arguments
  argument(argc, argv);

  pcie40reg_t pcie40;
  pcie40reg_t pcie40reg[48];
  
  // open pcie40 device driver for current process
  ecs_open( dev_slot , SLC_BAR );
  
  readregs(link_mask, pcie40reg);
  printf("statlink version %d (%d) / ", VERSION, MOD_DATE);
  statpice40(pcie40);
  statlink(link_mask, pcie40reg);
  
  // close pcie40 device driver for current process
  ecs_close( dev_slot , SLC_BAR) ;
  
  return 0 ;
}


