#!/bin/bash

CACHE_PATH=/tmp/pcie40_reload.cache
LOG_FILE=pcie40reload_log.txt

#ug`pcie40_reload.description`
# This command must be executed to reload the PCIe40 driver after an FPGA has been reprogrammed.
# _ The tool performs the following steps, in order:

USE_CACHE=0
FLUSH_CACHE=0
QUIET=0

#ug`pcie40_reload.synopsis`
# *pcie40_reload* [-c] [-f] [-q]

function usage {
  echo "pcie40_reload [-c] [-f] [-q]" >&2
  echo " -c    always read cache" >&2
  echo " -f    always overwrite cache" >&2
  echo " -q    suppress dmesg output" >&2
}

while getopts "macfqh" opt; do
  case $opt in

#ug`pcie40_reload.options`cache
# *-c*::
# Take the device list from the local cache, if it exists. By default the cache is used only if no PCIe40 interface appears under sysfs. With this option the cache contents are used instead. This can be useful in case a reload was issued too early (while one FPGA was still being reprogrammed) and the corresponding sysfs nodes do not exist anymore.
    c) USE_CACHE=1
      ;;
#ug`pcie40_reload.options`flush
# *-f*::
# Forces a cache update. By default the cache is updated only if additional entries have been detected compared to the cache contents. With this option the cache will be overwritten regardless.
    f) FLUSH_CACHE=1
      ;;
#ug`pcie40_reload.options`quiet
# *-q*::
# Suppress dmesg output. Without this option a partial dmesg log is printed for troubleshooting in case of errors.
    q) QUIET=1
      ;;
    h)
      usage
      exit 1
      ;;
    \?)
      echo "Invalid option: -${OPTARG}" >&2
      exit 1
      ;;
    :)
      echo "Option -${OPTARG} requires an argument" >&2
      exit 1
      ;;
  esac
done

if test -f "$LOG_FILE"; then
    cp "$LOG_FILE" "$LOG_FILE.1"
    rm "$LOG_FILE"
fi

devs=($(find /sys/devices/pci0000\:* -name pcie40_interface -exec cat {} \; -exec echo -n ' ' \; -execdir pwd \; | sort -nr | cut -f 2 -d ' ' 2>/dev/null))

if [ -z "${devs}" ]; then
  if [ -f "${CACHE_PATH}" ]; then
    echo "No PCIe40 devices appear in sysfs, using cache"
    echo "No PCIe40 devices appear in sysfs, using cache" >> "$LOG_FILE"
    devs=($(<${CACHE_PATH}))
  else
    echo "No devices detected! Assuming the FPGA was programmed correctly, you might have to reboot"
    echo "No devices detected! Assuming the FPGA was programmed correctly, you might have to reboot" >> "$LOG_FILE"
    exit 1
  fi
elif [ ${USE_CACHE} -eq 1 ]; then
  if [ -f "${CACHE_PATH}" ]; then
    echo "Using PCIe40 interface list from cache"
    echo "Using PCIe40 interface list from cache" >> "$LOG_FILE"
    devs=($(<${CACHE_PATH}))
  else
    echo "No PCIe40 interface cache present, if not all PCIe40 boards are detected you might have to reboot"
    echo "No PCIe40 interface cache present, if not all PCIe40 boards are detected you might have to reboot" >> "$LOG_FILE"
    exit 1
  fi
fi
if [ -f "${CACHE_PATH}" ]; then
  devs_cached=($(<${CACHE_PATH}))
  if [ ${FLUSH_CACHE} -eq 1 ] || [ ${#devs_cached[@]} -lt ${#devs[@]} ]; then
    echo "Updating cache"
    echo "Updating cache" >> "$LOG_FILE"
    echo "${devs}" > ${CACHE_PATH}
  fi
else
  echo "Creating cache" >> "$LOG_FILE"
  echo "${devs}" > ${CACHE_PATH}
fi

ok=1

pids_dma=
pids_bar0=
pids_bar2=

get_pids() {
# $1: file to check
# $2: variable to append pids to
  pids=`sudo lsof -t $1`
  for pid in $pids; do
    ps -p $pid --no-headers -o "  %U%p%c"
    eval "${2}+=\ ${pid}"
  done
}

if_id=${#devs[@]}

for dev in ${devs[@]}; do
  (( if_id -= 1 ))

  echo "PCIe40 interface ${if_id}:"
  echo "PCIe40 interface ${if_id}:" >> "$LOG_FILE"

  if [ $((if_id % 2)) -eq 0 ]; then
    if [ -e /dev/pcie40_${if_id}_bar0 ]; then
      echo " BAR0 used by:"
      get_pids /dev/pcie40_${if_id}_bar0 pids_bar0
      echo " BAR0 used by:" >> "$LOG_FILE"
      get_pids /dev/pcie40_${if_id}_bar0 pids_bar0 >> "$LOG_FILE"
    else
      echo " BAR0 missing!"
      echo " BAR0 missing!" >> "$LOG_FILE"
    fi
    if [ -e /dev/pcie40_${if_id}_bar2 ]; then
      echo " BAR2 used by:"
      get_pids /dev/pcie40_${if_id}_bar2 pids_bar2
      echo " BAR2 used by:" >> "$LOG_FILE"
      get_pids /dev/pcie40_${if_id}_bar2 pids_bar2 >> "$LOG_FILE"
    else
      echo " BAR2 missing!"
      echo " BAR2 missing!" >> "$LOG_FILE"
    fi
    if [ -e /dev/pcie40_${if_id}_dma ]; then
      echo " DMA used by:"
      get_pids /dev/pcie40_${if_id}_dma pids_dma
      echo " DMA used by:" >> "$LOG_FILE" 
      get_pids /dev/pcie40_${if_id}_dma pids_dma >> "$LOG_FILE"
    else
      echo " DMA missing!"
      echo " DMA missing!" >> "$LOG_FILE"
    fi
  fi
done

#ug`pcie40_reload.steps`1
# . Terminates any process currently using the PCIe40 boards to be reloaded

if [ -n "$pids_bar0" ] || [ -n "$pids_bar2" ]; then
  echo -n "Stopping processes... "
  echo -n "Stopping processes... " >> "$LOG_FILE"
  sudo kill -9 $pids_bar0 $pids_bar2 $pids_dma
  echo OK
  echo OK >> "$LOG_FILE"
fi

# Remove pll lock file for pll_status_small
sudo rm -f /tmp/pcie40_pll_lockfile.lck
# create a new one and give permission to all
sudo touch /tmp/pcie40_pll_lockfile.lck
sudo chmod 777 /tmp/pcie40_pll_lockfile.lck

#ug`pcie40_reload.steps`2
# . For every PCIe40 interface:
# .. Disconnects the interface via _sysfs_
# .. Requests a PCI rescan on the upstream device
# .. Checks if the driver detected successfully the interface

devs_removed=

echo "output of ps" >> "$LOG_FILE"
ps aux >> "$LOG_FILE"
echo "kernel output" >> "$LOG_FILE"
dmesg | grep -E "(P40:)|(P40DAQ:)" | tail -n 100 >> "$LOG_FILE"

sleep 1

sudo rmmod lhcb_pcie40 > /dev/null 2>&1
sudo rmmod pcie40
sudo insmod ${PCIE40_DRIVER_DIR}/pcie40.ko
for dev in ${devs[@]}; do
    if [ ! -d "${dev}" ]; then
	devs_removed="${dev} ${devs_removed}"
	continue;
    fi;
    echo "Removing device ${dev}"
    echo "Removing device ${dev}" >> "$LOG_FILE"
    if [ -f "${dev}/device" ] && [ $(< ${dev}/device) = 0xce40 ]; then
	if [ -e ${dev}/driver ]; then
            echo " Driver: OK"
            echo " Driver: OK" >> "$LOG_FILE"
	else
            echo " Driver: NO"
            echo " Driver: NO" >> "$LOG_FILE"
	fi;
	echo 1 | sudo tee ${dev}/remove > /dev/null
	devs_removed="${dev} ${devs_removed}"
    fi;
done

for dev in ${devs_removed}; do
    echo "Reloading device ${dev}"
    echo "Reloading device ${dev}" >> "$LOG_FILE"
    rp=`dirname ${dev}`
    echo 1 | sudo tee ${rp}/rescan > /dev/null
    if [ -e ${dev}/driver ]; then
	echo " Driver: OK"
	echo " Driver: OK" >> "$LOG_FILE"
    else
	echo " Driver: NO"
	echo " Driver: NO" >> "$LOG_FILE"
EOF
	ok=0
    fi;
done

if_id=${#devs[@]}

for dev in ${devs[@]}; do
  (( if_id -= 1 ))
  echo "PCIe40 interface ${if_id}:"
  echo "PCIe40 interface ${if_id}:" >> "$LOG_FILE"

  #ug`pcie40_reload.steps`3
  # . Checks the presence of the ECS registers

  if [ $((if_id % 2)) -eq 0 ]; then
    echo -n " BAR0: "
    echo -n " BAR0: " >> "$LOG_FILE"
     if [ -e /dev/pcie40_${if_id}_bar0 ]; then
       echo "OK"
       echo "OK" >> "$LOG_FILE"
     else
       echo "NO"
       echo "NO" >> "$LOG_FILE"
       ok=0
     fi
    echo -n " BAR2: "
    echo -n " BAR2: " >> "$LOG_FILE"
    if [ -e /dev/pcie40_${if_id}_bar2 ]; then
      echo "OK"
      echo "OK" >> "$LOG_FILE"
    else
      echo "NO"
      echo "NO" >> "$LOG_FILE"
      ok=0
    fi
  fi

  #ug`pcie40_reload.steps`4
  # . Checks the presence of the DMA controllers

  echo -n " DMA: "
  echo -n " DMA: " >> "$LOG_FILE"
  if [ -e /dev/pcie40_${if_id}_dma ]; then
    echo "OK"
    echo "OK" >> "$LOG_FILE"
  else
    echo "NO"
    echo "NO" >> "$LOG_FILE"
    ok=0
  fi
done

if [ "$ok" -eq 0 ] && [ ${QUIET} -eq 0 ]; then
  echo "Errors found, debug output follows:"
  echo "Errors found, debug output follows:" >> "$LOG_FILE"
  dmesg | grep -E "(P40:)|(P40DAQ:)" | tail -n 100
  dmesg | grep -E "(P40:)|(P40DAQ:)" | tail -n 100 >> "$LOG_FILE"
  exit 1
fi

echo "Change permissions" 
echo "Change permissions"  >> "$LOG_FILE"
sudo chmod a+rwx /dev/pcie40_0_*

#ug`pcie40_reload.exit`
# 0 on success, 1 if issues have been encountered. In the latter case the last diagnostic messages from the driver are also printed to standard output.
