// File for Belle 2 DMA functions for PCIe40: driver access
#include "dma_driverlib.h"

#include "pcie40_driverlib.h"

#include "pcie40_s_pages.h" 
#include "altera_dma_cmd.h" 
#include "pcie40_definitions.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sys/mman.h>
#include <errno.h>
#include <unistd.h>
#include <cstring>

static int dma_file[ MAX_DEV ] ;
static unsigned int *dpp[ LINK_NUMBER ][ S_PAGES ] ;

// **********************************************************************************
// Open device for DMA
// **********************************************************************************
int dma_open( int fd ) {
  char devname[32];
  struct dma_cmd G_dma_cmd;
  char cmdbuff[ 100 ] ;
  ssize_t f ;

  if ( p40_get_devname(fd, "dma", devname, sizeof(devname) ) < 0 ) {
    perror("p40_get_devname()");
    return -1;
  }

  f = open(devname, O_RDWR);

  if ( -1 == f ) {
    perror("open()");
    return -1;
  }

  // map spage addresses
  for ( int i=0 ; i<LINK_NUMBER ; i++ )
    for ( int j=0 ; j<S_PAGES ; j++ ) {
      
      //set spage for mapping;
      unsigned act_spage = ( i << 8 ) | j ;
      G_dma_cmd.cmd  = ALTERA_CMD_ACTSPAGE ;
      G_dma_cmd.buf  = cmdbuff ;
      *(int *)cmdbuff = act_spage ;
      write(f, &G_dma_cmd, 0) ;

      dpp[ i ][ j ] = (unsigned int *) mmap( NULL , 
					     S_PAGE_SIZE ,  
					     PROT_READ | PROT_WRITE ,
					     MAP_SHARED, 
					     f, 
					     0 ) ;  //map spage
      
      if (dpp[i][j] == MAP_FAILED) {
        perror( "mmap" );
        close ( f ) ; 
	      return -1 ;
      }
    }

  dma_file[ fd ] = f ;
  return f ;
}

// **************************************************************************
// Close device for DMA
// **************************************************************************
int dma_close( int fd ) {
  for ( int i=0 ; i<LINK_NUMBER ; i++ )
    for ( int j=0 ; j<S_PAGES ; j++ ) {
      munmap( dpp[ i ][ j ] , S_PAGE_SIZE ) ;
    }
  close( dma_file[ fd ] ) ; 
  dma_file[ fd ] = -1 ; 
  return 0 ;
}

// ***************************************************************************
// DMA initialize
// ***************************************************************************
int dma_initialize( int fd ) {
  ioctl( dma_file[ fd ] , ALTERA_IOCX_CLEAR);
  return 0 ;
}

// ***************************************************************************
// Reset the spages memory
// ***************************************************************************
int pcie40_resetSPages( int fd ) {
  for ( int i=0 ; i<LINK_NUMBER ; i++ )
    for ( int j=0 ; j<S_PAGES ; j++ )
      memset( dpp[i][j] , 0 , S_PAGE_SIZE ) ;
  return 0 ;
}

// **************************************************************************
// Get pointer to a super page
// **************************************************************************
unsigned int * pcie40_getSuperPagePointer( int link , int spage , int slot ) {
  if ( ( link >= LINK_NUMBER ) || ( spage >= S_PAGES ) || ( slot >= S_PAGE_SLOT_NMB ) ) 
    return 0 ;
  return &dpp[ link ][ spage ][ slot * S_PAGE_SLOT_SIZE/4 ] ;
}

// **************************************************************************
// Get pointer to a super page
// **************************************************************************
unsigned int * pcie40_getSuperPageCopy( int link , int spage , int slot ) {
  if ( ( link >= LINK_NUMBER ) || ( spage >= S_PAGES ) || ( slot >= S_PAGE_SLOT_NMB ) ) 
    return 0 ;
  unsigned int * pointer = &dpp[ link ][ spage ][ slot * S_PAGE_SLOT_SIZE/4 ] ;
  unsigned int * newdata = new unsigned int[ S_PAGE_SLOT_SIZE/4 ] ;
  memcpy( newdata , pointer , S_PAGE_SLOT_SIZE ) ;
  return newdata ;
}

// **************************************************************************
// Free the superpage when it has been read
// **************************************************************************
int pcie40_freeSuperPage( int fd , int spage ) {
  UAR uar ;
  uar.sr.spag_ready = spage ;
  ioctl(dma_file[ fd ], ALTERA_IOCX_RCONT,uar.arg);
  return 0 ;
}

// **************************************************************************
// Start the DMA
// **************************************************************************
int pcie40_start_dma(int fd) { 
  UAR uar ;
  ioctl( dma_file[ fd ] , ALTERA_IOCX_RSTART,&uar.arg); 
  return uar.sr.spag_ready ;
}
